/* hgc - Human Genome Click processor - gets called when user clicks
 * on something in human tracks display. */
#include "common.h"
#include "hCommon.h"
#include "portable.h"
#include "dystring.h"
#include "cheapcgi.h"
#include "htmshell.h"
#include "jksql.h"
#include "dnautil.h"
#include "dnaseq.h"
#include "fa.h"
#include "fuzzyFind.h"
#include "hdb.h"
#include "hgRelate.h"
#include "agpFrag.h"
#include "ctgPos.h"
#include "clonePos.h"
#include "rmskOut.h"
#include "xenalign.h"
#include "isochores.h"
#include "simpleRepeat.h"
#include "cpgIsland.h"
#include "genePred.h"
#include "pepPred.h"
#include "genomicDups.h"
#include "psl.h"
#include "wabAli.h"

char *seqName;		/* Name of sequence we're working on. */
int winStart, winEnd;   /* Bounds of sequence. */

char *entrezScript = "http://www.ncbi.nlm.nih.gov/htbin-post/Entrez/query?form=4";

void printEntrezNucleotideUrl(FILE *f, char *accession)
/* Print URL for Entrez browser on a nucleotide. */
{
fprintf(f, "\"%s&db=n&term=%s\"", entrezScript, accession);
}

void hgcAnchorSomewhere(char *group, char *item, char *other, char *chrom)
/* Generate an anchor that calls click processing program with item and other parameters. */
{
printf("<A HREF=\"../cgi-bin/hgc?g=%s&i=%s&c=%s&l=%d&r=%d%s\">",
	group, item, chrom, winStart, winEnd, other);
}

void hgcAnchor(char *group, char *item, char *other)
/* Generate an anchor that calls click processing program with item and other parameters. */
{
hgcAnchorSomewhere(group, item, other, seqName);
}

void doGetDna()
/* Get dna in window. */
{
struct dnaSeq *seq;
char name[256];
sprintf(name, "%s:%d-%d", seqName, winStart, winEnd);
htmlStart(name);
seq = hDnaFromSeq(seqName, winStart, winEnd, dnaMixed);
printf("<TT><PRE>");
faWriteNext(stdout, name, seq->dna, seq->size);
printf("</TT></PRE>");
freeDnaSeq(&seq);
}

void medlineLinkedLine(char *title, char *text, char *search)
/* Produce something that shows up on the browser as
 *     TITLE: value
 * with the value hyperlinked to medline. */
{
char *encoded = cgiEncode(search);

printf("<B>%s:</B> ", title);
if (sameWord(text, "n/a"))
    printf("n/a<BR>\n");
else
    printf("<A HREF=\"%s&db=m&term=%s\" TARGET=_blank>%s</A><BR>", entrezScript, 
        encoded, text);
freeMem(encoded);
}

void appendAuthor(struct dyString *dy, char *gbAuthor, int len)
/* Convert from  Kent,W.J. to Kent WJ and append to dy.
 * gbAuthor gets eaten in the process. */
{
char buf[256];

if (len >= sizeof(buf))
    warn("author %s too long to process", gbAuthor);
else
    {
    memcpy(buf, gbAuthor, len);
    buf[len] = 0;
    stripChar(buf, '.');
    subChar(buf, ',' , ' ');
    dyStringAppend(dy, buf);
    dyStringAppend(dy, " ");
    }
}

void gbToEntrezAuthor(char *authors, struct dyString *dy)
/* Convert from Genbank author format:
 *      Kent,W.J., Haussler,D. and Zahler,A.M.
 * to Entrez search format:
 *      Kent WJ,Haussler D,Zahler AM
 */
{
char buf[256];
char *s = authors, *e;

/* Parse first authors, which will be terminated by '.,' */
while ((e = strstr(s, ".,i ")) != NULL)
    {
    int len = e - s + 1;
    appendAuthor(dy, s, len);
    s += len+2;
    }
if ((e = strstr(s, " and")) != NULL)
    {
    int len = e - s;
    appendAuthor(dy, s, len);
    s += len+4;
    }
if ((s = skipLeadingSpaces(s)) != NULL && s[0] != 0)
    {
    int len = strlen(s);
    appendAuthor(dy, s, len);
    }
}

void printRnaSpecs(char *acc)
/* Print auxiliarry info on RNA. */
{
struct dyString *dy = newDyString(1024);
struct sqlConnection *conn = hgAllocConn();
struct sqlResult *sr;
char **row;
char *type,*direction,*source,*organism,*library,*clone,*sex,*tissue,
     *development,*cell,*cds,*description, *author,*geneName,*productName;
int seqSize,fileSize;
long fileOffset;
char *date;
char *ext_file;

/* This sort of query and having to keep things in sync between
 * the first clause of the select, the from clause, the where
 * clause, and the results in the row ... is really tedious.
 * One of my main motivations for going to a more object
 * based rather than pure relational approach in general,
 * and writing 'autoSql' to help support this.  However
 * the pure relational approach wins for pure search speed,
 * and these RNA fields are searched.  So it looks like
 * the code below stays.  Be really careful when you modify
 * it. */
dyStringAppend(dy,
  "select mrna.type,mrna.direction,"
  	 "source.name,organism.name,library.name,mrnaClone.name,"
         "sex.name,tissue.name,development.name,cell.name,cds.name,"
	 "description.name,author.name,geneName.name,productName.name,"
	 "seq.size,seq.gb_date,seq.extFile,seq.file_offset,seq.file_size "
  "from mrna,seq,source,organism,library,mrnaClone,sex,tissue,"
        "development,cell,cds,description,author,geneName,productName ");
dyStringPrintf(dy,
  "where mrna.acc = '%s' and mrna.id = seq.id ", acc);
dyStringAppend(dy,
    "and mrna.source = source.id and mrna.organism = organism.id "
    "and mrna.library = library.id and mrna.mrnaClone = mrnaClone.id "
    "and mrna.sex = sex.id and mrna.tissue = tissue.id "
    "and mrna.development = development.id and mrna.cell = cell.id "
    "and mrna.cds = cds.id and mrna.description = description.id "
    "and mrna.author = author.id and mrna.geneName = geneName.id "
    "and mrna.productName = productName.id");
sr = sqlMustGetResult(conn, dy->string);
row = sqlNextRow(sr);
type=row[0];direction=row[1];source=row[2];organism=row[3];library=row[4];clone=row[5];
sex=row[6];tissue=row[7];development=row[8];cell=row[9];cds=row[10];description=row[11];
author=row[12];geneName=row[13];productName=row[14];
seqSize = sqlUnsigned(row[15]);
date = row[16];
ext_file = row[17];
fileOffset=sqlUnsigned(row[18]);
fileSize=sqlUnsigned(row[19]);

/* Now we have all the info out of the database and into nicely named
 * local variables.  There's still a few hoops to jump through to 
 * format this prettily on the web with hyperlinks to NCBI. */
printf("<H2>Information on %s <A HREF=", type);
printEntrezNucleotideUrl(stdout, acc);
printf(" TARGET=_blank>%s</A></H2>\n", acc);

printf("<B>description:</B> %s<BR>\n", description);

medlineLinkedLine("gene", geneName, geneName);
medlineLinkedLine("product", productName, productName);
dyStringClear(dy);
gbToEntrezAuthor(author, dy);
medlineLinkedLine("author", author, dy->string);
printf("<B>organism:</B> %s<BR>\n", organism);
printf("<B>tissue:</B> %s<BR>\n", tissue);
printf("<B>development stage:</B> %s<BR>\n", development);
printf("<B>cell type:</B> %s<BR>\n", cell);
printf("<B>sex:</B> %s<BR>\n", sex);
printf("<B>library:</B> %s<BR>\n", library);
printf("<B>clone:</B> %s<BR>\n", clone);
if (direction[0] != '0') printf("<B>read direction:</B> %s'<BR>\n", direction);
printf("<B>cds:</B> %s<BR>\n", cds);

sqlFreeResult(&sr);
freeDyString(&dy);
hgFreeConn(&conn);
}

void doHgRna(char *acc, boolean isEst)
/* Click on an individual RNA. */
{
char query[256];
struct sqlConnection *conn = hAllocConn();
struct sqlResult *sr = NULL;
char **row;
struct linkedFeatures *lfList = NULL, *lf;
char *type = (isEst ? "est" : "mrna");
int start = cgiInt("o");
struct psl *pslList = NULL, *psl;
int aliCount;
char otherString[32];
boolean same;

/* Print non-sequence info. */
htmlStart(acc);
printRnaSpecs(acc);

/* Get alignment info. */
sprintf(query, "select * from all_%s where qName = '%s'",
    type, acc);
sr = sqlGetResult(conn, query);
while ((row = sqlNextRow(sr)) != NULL)
    {
    psl = pslLoad(row);
    slAddHead(&pslList, psl);
    }
sqlFreeResult(&sr);
slReverse(&pslList);

/* Print alignments. */
htmlHorizontalLine();
printf("<H3>%s/Genomic Alignments</H3>", (isEst ? "EST" : "mRNA"));
aliCount = slCount(pslList);
if (aliCount > 1)
    printf("The alignment you clicked on is first in the table below.<BR>\n");

printf("<TT><PRE>");
printf(" SIZE IDENTITY  CHROMOSOME   START     END       cDNA   START  END  TOTAL\n");
printf("-------------------------------------------------------------------------\n");
for (same = 1; same >= 0; same -= 1)
    {
    for (psl = pslList; psl != NULL; psl = psl->next)
	{
	if (same ^ (psl->tStart != start))
	    {
	    sprintf(otherString, "&o=%d&type=%s", psl->tStart, type);
	    hgcAnchorSomewhere("htcCdnaAli", acc, otherString, psl->tName);
	    printf("%5d  %5.1f%%  %9s  %9d %9d  %8s %5d %5d %5d</A>",
		psl->match + psl->misMatch + psl->repMatch + psl->nCount,
		100.0 - pslCalcMilliBad(psl, TRUE) * 0.1,
		skipChr(psl->tName), psl->tStart + 1, psl->tEnd,
		psl->qName, psl->qStart+1, psl->qEnd, psl->qSize);
	    printf("\n");
	    }
	}
    }
}

void doHgGold(char *fragName)
/* Click on a fragment of golden path. */
{
struct sqlConnection *conn = hAllocConn();
char query[256];
struct sqlResult *sr;
char **row;
struct agpFrag frag;
struct dnaSeq *seq;
char dnaName[256];

htmlStart(fragName);
sprintf(query, "select * from %s_gold where frag = '%s'", seqName, fragName);
sr = sqlMustGetResult(conn, query);
row = sqlNextRow(sr);
agpFragStaticLoad(row, &frag);

printf("Fragment %s bases %d-%d strand %s makes up draft sequence of bases %d-%d of chromosome %s<BR>\n",
	frag.frag, frag.fragStart+1, frag.fragEnd, frag.strand,
	frag.chromStart+1, frag.chromEnd, skipChr(frag.chrom));
htmlHorizontalLine();

sprintf(dnaName, "%s:%d-%d", frag.chrom, frag.chromStart, frag.chromEnd);
seq = hDnaFromSeq(frag.chrom, frag.chromStart, frag.chromEnd, dnaMixed);
printf("<TT><PRE>");
faWriteNext(stdout, dnaName, seq->dna, seq->size);
printf("</TT></PRE>");

freeDnaSeq(&seq);
sqlFreeResult(&sr);
hFreeConn(&conn);
}

void doHgGap(char *gapType)
/* Print a teeny bit of info about a gap. */
{
char *words[2];
int wordCount;

htmlStart("Gap in Sequence");
chopByChar(gapType, '_', words, ArraySize(words));
printf("This is a gap between %ss.<BR>", words[0]);
if (sameString(words[1], "no"))
    printf("This gap is not bridged by an mRNA, EST or BAC end pair.\n"); 
else
    printf("This gap is bridged by mRNA, EST or BAC end pairs.\n");
}

void doHgContig(char *ctgName)
/* Click on a contig. */
{
struct sqlConnection *conn = hAllocConn();
char query[256];
struct sqlResult *sr;
char **row;
struct ctgPos *ctg;
int cloneCount;

htmlStart(ctgName);
sprintf(query, "select * from ctgPos where contig = '%s'", ctgName);
sr = sqlMustGetResult(conn, query);
row = sqlNextRow(sr);
ctg = ctgPosLoad(row);
sqlFreeResult(&sr);
sprintf(query, 
    "select count(*) from clonePos where chrom = '%s' and chromEnd >= %d and chromStart <= %d",
    ctg->chrom, ctg->chromStart, ctg->chromEnd);
cloneCount = sqlQuickNum(conn, query);

printf("Fingerprint Map Contig %s includes %d clones and makes up bases %d-%d of the draft sequence of chromosome %s.<BR>\n",
     ctgName, cloneCount, ctg->chromStart+1, ctg->chromEnd, skipChr(ctg->chrom));
printf("Go back and type '%s' in the position field to bring the entire contig into the browser window.",
     ctgName);
hFreeConn(&conn);
}

char *cloneStageName(char *stage)
/* Expand P/D/F. */
{
switch (stage[0])
   {
   case 'P':
      return "predraft (less than 4x coverage shotgun)";
   case 'D':
      return "draft (at least 4x coverage shotgun)";
   case 'F':
      return "finished";
   }
}

void doHgCover(char *cloneName)
/* Respond to click on clone. */
{
struct sqlConnection *conn = hAllocConn();
char query[256];
struct sqlResult *sr;
char **row;
struct clonePos *clone;
int fragCount;

htmlStart(cloneName);
sprintf(query, "select * from clonePos where name = '%s'", cloneName);
sr = sqlMustGetResult(conn, query);
row = sqlNextRow(sr);
clone = clonePosLoad(row);
sqlFreeResult(&sr);

sprintf(query, 
    "select count(*) from %s_gl where end >= %d and start <= %d and frag like '%s%%'",
    clone->chrom, clone->chromStart, clone->chromEnd, clone->name);
fragCount = sqlQuickNum(conn, query);


printf("<H2>Information on <A HREF=");
printEntrezNucleotideUrl(stdout, cloneName);
printf(" TARGET=_blank>%s</A></H2>\n", cloneName);
printf("<B>status:</B> %s<BR>\n", cloneStageName(clone->stage));
printf("<B>fragments:</B> %d<BR>\n", fragCount);
printf("<B>size:</B> %d bases<BR>\n", clone->seqSize);
printf("<B>chromosome:</B> %s<BR>\n", skipChr(clone->chrom));
printf("<BR>\n");
hgcAnchor("htcCloneSeq", cloneName, "");
printf("Fasta format sequence</A><BR>\n");
hFreeConn(&conn);
}

void doHgClone(char *fragName)
/* Handle click on a clone. */
{
char cloneName[128];
fragToCloneVerName(fragName, cloneName);
doHgCover(cloneName);
}

void htcCloneSeq(char *cloneName)
/* Fetch sequence as fasta file. */
{
struct sqlConnection *conn = hAllocConn();
char query[256];
struct sqlResult *sr;
char **row;
struct clonePos *clone;
struct dnaSeq *seqList, *seq;

htmlStart(cloneName);
sprintf(query, "select * from clonePos where name = '%s'", cloneName);
sr = sqlMustGetResult(conn, query);
row = sqlNextRow(sr);
clone = clonePosLoad(row);
sqlFreeResult(&sr);

seqList = faReadAllDna(clone->faFile);
printf("<PRE><TT>");
for (seq = seqList; seq != NULL; seq = seq->next)
    {
    faWriteNext(stdout, seq->name, seq->dna, seq->size);
    }
hFreeConn(&conn);
}

void htcCdnaAli(char *acc)
/* Show alignment for accession. */
{
struct tempName indexTn, bodyTn;
FILE *index, *body;
struct dnaSeq *rnaSeq;
struct dnaSeq *dnaSeq;
DNA *rna;
int dnaSize,rnaSize;
boolean isRc = FALSE;
char *type = cgiString("type");
int start = cgiInt("o");
char query[256];
struct sqlConnection *conn = hAllocConn();
struct sqlResult *sr;
char **row;
struct psl *psl;
struct ffAli *ffAli, *ff;
int tStart, tEnd;
int lastEnd = 0;
int blockCount;
int i;
char title[256];

/* Print start of HTML. */
puts("Content-Type:text/html\n");
puts("<HTML>");
printf("<HEAD>\n<TITLE>%s vs Genomic</TITLE>\n</HEAD>\n\n", acc);

makeTempName(&indexTn, "index", ".html");
makeTempName(&bodyTn, "body", ".html");

/* Look up alignments in database */
sprintf(query, "select * from %s_%s where qName = '%s' and tStart=%d",
    seqName, type, acc, start);
sr = sqlGetResult(conn, query);
if ((row = sqlNextRow(sr)) == NULL)
    errAbort("Couldn't find alignment for %s at %d", acc, start);
psl = pslLoad(row);

/* Get RNA and DNA sequence.  Save a mixed case copy of DNA, make
 * all lower case for fuzzyFinder. */
rnaSeq = hgRnaSeq(acc);
rnaSize = rnaSeq->size;
rna = rnaSeq->dna;
tStart = psl->tStart - 100;
if (tStart < 0) tStart = 0;
tEnd  = psl->tEnd + 100;
if (tEnd > psl->tSize) tEnd = psl->tSize;
dnaSeq = hDnaFromSeq(seqName, tStart, tEnd, dnaMixed);
freez(&dnaSeq->name);
dnaSeq->name = cloneString(psl->tName);


/* Convert psl alignment to ffAli. */
if (psl->strand[0] == '-')
    {
    isRc = TRUE;
    reverseComplement(rna, rnaSize);
    }
ffAli = pslToFfAli(psl, rnaSeq, dnaSeq, tStart);
if (isRc)
    reverseComplement(rna, rnaSize);

/* Write body of alignment. */
body = mustOpen(bodyTn.forCgi, "w");
htmStart(body, acc);
fprintf(body, "<H2>Alignment of %s and %s:%d-%d</H2>\n", acc, psl->tName, psl->tStart, psl->tEnd);
fprintf(body, "Click on links in the frame to left to navigate through alignment.\n");
blockCount = ffShAliPart(body, ffAli, acc, rna, rnaSize, 0, 
	dnaSeq->name, dnaSeq->dna, dnaSeq->size, tStart, 
	8, isRc, FALSE, TRUE, TRUE, TRUE);
fclose(body);
chmod(bodyTn.forCgi, 0666);

/* Write index. */
index = mustOpen(indexTn.forCgi, "w");
htmStart(index, acc);
fprintf(index, "<H3>%s</H3>", acc);
fprintf(index, "<A HREF=\"%s#cDNA\" TARGET=\"body\">cDNA</A><BR>\n", bodyTn.forCgi);
fprintf(index, "<A HREF=\"%s#genomic\" TARGET=\"body\">genomic</A><BR>\n", bodyTn.forCgi);
for (i=1; i<=blockCount; ++i)
    {
    fprintf(index, "<A HREF=\"%s#%d\" TARGET=\"body\">block%d</A><BR>\n",
         bodyTn.forCgi, i, i);
    }
fprintf(index, "<A HREF=\"%s#ali\" TARGET=\"body\">together</A><BR>\n", bodyTn.forCgi);
fclose(index);
chmod(indexTn.forCgi, 0666);

/* Write (to stdout) the main html page containing just the frame info. */
puts("<FRAMESET COLS = \"13%,87% \" >");
printf("  <FRAME SRC=\"%s\" NAME=\"index\" RESIZE>\n", indexTn.forCgi);
printf("  <FRAME SRC=\"%s\" NAME=\"body\" RESIZE>\n", bodyTn.forCgi);
puts("</FRAMESET>");
puts("<NOFRAMES><BODY></BODY></NOFRAMES>");
}

void writeMatches(FILE *f, char *a, char *b, int count)
/* Write a | where a and b agree, a ' ' elsewhere. */
{
int i;
for (i=0; i<count; ++i)
    {
    if (a[i] == b[i])
        fputc('|', f);
    else
        fputc(' ', f);
    }
}

void fetchAndShowWaba(char *table, char *name)
/* Fetch and display waba alignment. */
{
struct sqlConnection *conn = hAllocConn();
struct sqlResult *sr;
char **row;
char query[256];
int start = cgiInt("o");
struct wabAli *wa;
int qOffset;
char strand = '+';

sprintf(query, "select * from %s where query = '%s' and chrom = '%s' and chromStart = %d",
	table, name, seqName, start);
sr = sqlGetResult(conn, query);
if ((row = sqlNextRow(sr)) == NULL)
    errAbort("Sorry, couldn't find alignment of %s at %d of %s in database",
    	name, start, seqName);
wa = wabAliLoad(row);
printf("<PRE><TT>");
qOffset = wa->qStart;
if (wa->strand[0] == '-')
    {
    strand = '-';
    qOffset = wa->qEnd;
    }
xenShowAli(wa->qSym, wa->tSym, wa->hSym, wa->symCount, stdout,
    qOffset, wa->chromStart, strand, '+', 60);
printf("</PRE></TT>");

wabAliFree(&wa);
hFreeConn(&conn);
}

void doHgTet(char *name)
/* Do thing with tet track. */
{
htmlStart("Tetraodon Alignment");
printf("Alignment between tetraodon sequence %s (above) and human chromosome %s (below)\n",
    name, skipChr(seqName));
fetchAndShowWaba("waba_tet", name);
}

void doHgRepeat(char *repeat)
/* Do click on a repeat track. */
{
htmlStart("Repeat");
if (cgiVarExists("o"))
    {
    struct sqlConnection *conn = hAllocConn();
    struct sqlResult *sr;
    char **row;
    struct rmskOut *ro;
    char query[256];
    int start = cgiInt("o");
    sprintf(query, "select * from %s_rmsk where  repName = '%s' and genoName = '%s' and genoStart = %d",
	    seqName, repeat, seqName, start);
    sr = sqlGetResult(conn, query);
    printf("<H3>RepeatMasker Information</H3>\n");
    while ((row = sqlNextRow(sr)) != NULL)
	{
	ro = rmskOutLoad(row);
	printf("<B>Name:</B> %s<BR>\n", ro->repName);
	printf("<B>Family:</B> %s<BR>\n", ro->repFamily);
	printf("<B>Class:</B> %s<BR>\n", ro->repClass);
	printf("<B>SW Score:</B> %d<BR>\n", ro->swScore);
	printf("<B>Divergence:</B> %3.1f%%<BR>\n", 0.1 * ro->milliDiv);
	printf("<B>Deletions:</B>  %3.1f%%<BR>\n", 0.1 * ro->milliDel);
	printf("<B>Insertions:</B> %3.1f%%<BR>\n", 0.1 * ro->milliIns);
	printf("<B>Begin in repeat:</B> %d<BR>\n", ro->repStart);
	printf("<B>End in repeat:</B> %d<BR>\n", ro->repEnd);
	printf("<B>Left in repeat:</B> %d<BR>\n", ro->repLeft);
	printf("<B>Chromosome:</B> %s<BR>\n", skipChr(ro->genoName));
	printf("<B>Begin in chromosome:</B> %d<BR>\n", ro->genoStart);
	printf("<B>End in chromosome:</B> %d<BR>\n", ro->genoEnd);
	printf("<B>Strand:</B> %s<BR>\n", ro->strand);
	printf("<BR>\n");
	}
    hFreeConn(&conn);
    }
else
    {
    if (sameString(repeat, "SINE"))
	printf("This track contains the small interspersed (SINE) class of repeats , which includes ALUs.<BR>\n");
    else if (sameString(repeat, "LINE"))
        printf("This track contains the long interspersed (LINE) class of repeats<BR>\n");
    else if (sameString(repeat, "LTR"))
        printf("This track contains the LTR (long terminal repeat) class of repeats which includes retroposons<BR>\n");
    else
        printf("This track contains the %s class of repeats<BR>\n", repeat);
    printf("Click right on top of an individual repeat for more information on that repeat<BR>\n");
    }
}

void doHgIsochore(char *item)
/* do click on isochore track. */
{
htmlStart("Isochore Info");
printf("<H2>Isochore Information</H2>\n");
if (cgiVarExists("o"))
    {
    struct isochores *iso;
    struct sqlConnection *conn = hAllocConn();
    struct sqlResult *sr;
    char **row;
    char query[256];
    int start = cgiInt("o");
    sprintf(query, "select * from isochores where  name = '%s' and chrom = '%s' and chromStart = %d",
	    item, seqName, start);
    sr = sqlGetResult(conn, query);
    while ((row = sqlNextRow(sr)) != NULL)
	{
	iso = isochoresLoad(row);
	printf("<B>Type:</B> %s<BR>\n", iso->name);
	printf("<B>GC Content:</B> %3.1f%%<BR>\n", 0.1*iso->gcPpt);
	printf("<B>Chromosome:</B> %s<BR>\n", skipChr(iso->chrom));
	printf("<B>Begin in chromosome:</B> %d<BR>\n", iso->chromStart);
	printf("<B>End in chromosome:</B> %d<BR>\n", iso->chromEnd);
	printf("<B>Size:</B> %d<BR>\n", iso->chromEnd - iso->chromStart);
	printf("<BR>\n");
	isochoresFree(&iso);
	}
    hFreeConn(&conn);
    }
htmlHorizontalLine();
puts(
    "<h3>What's an Isochore</h3>"
    "<P>Isochores describe a region of a chromosome where the CG-content is"
    " either higher or lower than the whole genome average (42%).  A CG-rich"
    " isochore is given a dark color, while a CG-poor isochore is a light"
    " color.  </P>"
    ""
    "<P>Isochores were determined by first calculating the CG-content of 100,000bp"
    " windows across the genome.  These windows were either labeled H or L"
    " depending on whether the window contained a higher or lower GC-content"
    " than average.  A two-state HMM was created in which one state represented"
    " GC-rich regions, and the other GC-poor.  It was trained using the first 12"
    " chromosomes. The trained HMM was used to generate traces over all chromosomes."
    " These traces define the boundaries of the isochores,"
    " and their type (GC-rich or AT-rich).</P>"
    );
}

void doSimpleRepeat(char *item)
/* Print info on simple repeat. */
{
htmlStart("Simple Repeat Info");
printf("<H2>Simple Tandem Repeat Information</H2>\n");
if (cgiVarExists("o"))
    {
    struct simpleRepeat *rep;
    struct sqlConnection *conn = hAllocConn();
    struct sqlResult *sr;
    char **row;
    char query[256];
    int start = cgiInt("o");
    sprintf(query, "select * from simpleRepeat where  name = '%s' and chrom = '%s' and chromStart = %d",
	    item, seqName, start);
    sr = sqlGetResult(conn, query);
    while ((row = sqlNextRow(sr)) != NULL)
	{
	rep = simpleRepeatLoad(row);
	printf("<B>Chromosome:</B> %s<BR>\n", skipChr(rep->chrom));
	printf("<B>Begin in chromosome:</B> %d<BR>\n", rep->chromStart);
	printf("<B>End in chromosome:</B> %d<BR>\n", rep->chromEnd);
	printf("<B>Size:</B> %d<BR>\n", rep->chromEnd - rep->chromStart);
	printf("<B>Period:</B> %d<BR>\n", rep->period);
	printf("<B>Copies:</B> %4.1f<BR>\n", rep->copyNum);
	printf("<B>Consensus size:</B> %d<BR>\n", rep->consensusSize);
	printf("<B>Match Percentage:</B> %d%%<BR>\n", rep->perMatch);
	printf("<B>Insert/Delete Percentage:</B> %d%%<BR>\n", rep->perIndel);
	printf("<B>Score:</B> %d<BR>\n", rep->score);
	printf("<B>Entropy:</B> %4.3f<BR>\n", rep->entropy);
	printf("<B>Sequence:</B> %s<BR>\n", rep->sequence);
	printf("<BR>\n");
	simpleRepeatFree(&rep);
	}
    hFreeConn(&conn);
    }
else
    {
    puts("<P>Click directly on a repeat for specific information on that repeat</P>");
    }
htmlHorizontalLine();
puts("<P>The Simple Tandem Repeats were calculated by James Durbin</P>");
}

void doCpgIsland(char *item)
/* Print info on CpG Island. */
{
htmlStart("CpG Island Info");
printf("<H2>CpG Island Info</H2>\n");
if (cgiVarExists("o"))
    {
    struct cpgIsland *island;
    struct sqlConnection *conn = hAllocConn();
    struct sqlResult *sr;
    char **row;
    char query[256];
    int start = cgiInt("o");
    sprintf(query, "select * from cpgIsland where  name = '%s' and chrom = '%s' and chromStart = %d",
	    item, seqName, start);
    sr = sqlGetResult(conn, query);
    while ((row = sqlNextRow(sr)) != NULL)
	{
	island = cpgIslandLoad(row);
	printf("<B>Chromosome:</B> %s<BR>\n", skipChr(island->chrom));
	printf("<B>Begin in chromosome:</B> %d<BR>\n", island->chromStart);
	printf("<B>End in chromosome:</B> %d<BR>\n", island->chromEnd);
	printf("<B>Size:</B> %d<BR>\n", island->chromEnd - island->chromStart);
	printf("<B>CpG count:</B> %d<BR>\n", island->cpgNum);
	printf("<B>C count plus G count:</B> %d<BR>\n", island->gcNum);
	printf("<B>Percentage CpG:</B> %d%%<BR>\n", island->perCpg);
	printf("<B>Percentage C or G:</B> %d%%<BR>\n", island->perGc);
	printf("<BR>\n");
	cpgIslandFree(&island);
	}
    hFreeConn(&conn);
    }
else
    {
    puts("<P>Click directly on a CpG island for specific information on that island</P>");
    }
htmlHorizontalLine();
puts("<P>CpG islands are associated with genes, particularly housekeeping genes, in vertebrates. "
 "CpG islands are particularly common near transcription start sites, and may be associated "
 "with promoter regions.  The CpG islands displayed in this browser are all at least 50 "
 "bases long, and have a GC content of at least 50%.</P>");
}

void printLines(FILE *f, char *s, int lineSize)
/* Print s, lineSize characters (or less) per line. */
{
int len = strlen(s);
int start, left;
int oneSize;

printf("<TT><PRE>");
for (start = 0; start < len; start += lineSize)
    {
    oneSize = len - start;
    if (oneSize > lineSize)
        oneSize = lineSize;
    mustWrite(f, s+start, oneSize);
    fputc('\n', f);
    }
if (start != len)
    fputc('\n', f);
printf("</PRE></TT>");
}

void showProteinPrediction(char *geneName, char *table)
/* Fetch and display protein prediction. */
{
char query[512];
struct sqlConnection *conn = hAllocConn();
struct sqlResult *sr;
char **row;
struct pepPred *pp = NULL;

sprintf(query, "select * from %s where name = '%s'", table, geneName);
sr = sqlGetResult(conn, query);
if ((row = sqlNextRow(sr)) != NULL)
    pp = pepPredLoad(row);
sqlFreeResult(&sr);
if (pp != NULL)
    {
    printf("<H3>Translated Protein</H3>\n");
    printLines(stdout, pp->seq, 50);
    }
hFreeConn(&conn);
}

void doKnownGene(char *geneName)
/* Handle click on known gene track. */
{
htmlStart("Known Gene");
printf("<H2>Known Gene %s</H2>\n", geneName);
showProteinPrediction(geneName, "genieKnownPep");
htmlHorizontalLine();
puts(
   "<P>Known genes are derived from the "
   "<A HREF = \"http://www.ncbi.nlm.nih.gov/LocusLink/\" TARGET=_blank>"
   "LocusLink</A> database.  These were mapped to the draft "
   "human genome using <A HREF = \"http://www.cse.ucsc.edu/~kent\" TARGET=_blank>"
   "Jim Kent's</A> PatSpace software, and further processed by "
   "<A HREF = \"http://www.neomorphic.com\" TARGET=_blank>Neomorphic</A> to cope "
   "with sequencing errors in the draft.</P>\n"
   "<P>The treatment of known genes in the browser is quite "
   "preliminary.  In particular the current names for the genes "
   "are a temporary measure.  We hope to adopt HUGO names, and "
   "to provide links into LocusLink in the near future.</P>"
   "<P>More information is available with the mRNA associated "
   "with this gene</P>");
}

void doGeniePred(char *geneName)
/* Handle click on Genie prediction track. */
{
htmlStart("Genie Gene Prediction");
printf("<H2>Genie Gene Prediction %s</H2>\n", geneName);
showProteinPrediction(geneName, "genieAltPep");
htmlHorizontalLine();
puts(
   "<P>Genie predictions are based on "
   "<A HREF = \"http://www.neomorphic.com\" TARGET=_blank>Neomorphic's</A> "
   "Genie gene finding software.  Genie is a generalized HMM "
   "which accepts constraints based on mRNA and EST data. </P>"
   "<P>The treatment of predicted genes in the browser is still "
   "preliminary.  In particular gene names are not kept stable "
   "between versions of the draft human genome.");
}

void doEnsPred(char *geneName)
/* HAndle click on Ensembl gene track. */
{
htmlStart("Ensembl Prediction");
printf("<H2>Ensembl Prediction %s</H2>\n", geneName);
showProteinPrediction(geneName, "ensPep");
htmlHorizontalLine();
puts(
   "<P>This gene prediction was created by "
   "<A HREF = \"http://www.ensembl.org\" TARGET=_blank>Ensembl</A>."
   "<P>The treatment of predicted genes in the browser is still "
   "preliminary.  In particular gene names are not kept stable "
   "between versions of the draft human genome.");
}

void parseChromPointPos(char *pos, char *retChrom, int *retPos)
/* Parse out chrN:123 into chrN and 123. */
{
char *s, *e;
int len;
e = strchr(pos, ':');
if (e == NULL)
   errAbort("No : in chromosome point position %s", pos);
len = e - pos;
memcpy(retChrom, pos, len);
retChrom[len] = 0;
s = e+1;
*retPos = atoi(s);
}

void doGenomicDups(char *dupName)
/* Handle click on genomic dup track. */
{
struct genomicDups dup;
char query[512];
struct sqlConnection *conn = hAllocConn();
struct sqlResult *sr;
char **row;
char oChrom[64];
int oStart;



htmlStart("Genomic Duplications");
printf("<H2>Genomic Duplication</H2>\n");
if (cgiVarExists("o"))
    {
    int start = cgiInt("o");
    parseChromPointPos(dupName, oChrom, &oStart);

    sprintf(query, "select * from genomicDups where chrom = '%s' and chromStart = %d "
		   "and otherChrom = '%s' and otherStart = %d",
		   seqName, start, oChrom, oStart);
    sr = sqlGetResult(conn, query);
    while (row = sqlNextRow(sr))
	{
	genomicDupsStaticLoad(row, &dup);
	printf("<B>First Position:</B> %s:%d-%d<BR>\n",
	   dup.chrom, dup.chromStart, dup.chromEnd);
	printf("<B>Second Position:</B> %s:%d-%d<BR>\n",
	   dup.otherChrom, dup.otherStart, dup.otherEnd);
	printf("<B>Relative orientation:</B> %s<BR>\n", dup.strand);
	printf("<B>Percent identity:</B> %3.1f%%<BR>\n", 0.1*dup.score);
	printf("<B>Size:</B> %d<BR>\n", dup.alignB);
	printf("<B>Bases matching:</B> %d<BR>\n", dup.matchB);
	printf("<B>Bases not matching:</B> %d<BR>\n", dup.mismatchB);
	htmlHorizontalLine();
	}
    }
else
    {
    puts("<P>Click directly on a repeat for specific information on that repeat</P>");
    }
puts("This region was detected as a duplication within the golden path. "
     "Regions with more than 1000 bases of non-RepeatMasked sequence and "
     "greater than 90% pairwise similarity are included in this track. "
     "Duplications of 99% or greater similarity are shown as orange. "
     "Duplications of 98% - 99% similarity are shown as yellow. "
     "Duplications of 90% - 98% similarity are shown as shades of gray. "
     "The data was provided by <A HREF=\"mailto:jab@cwru.edu\">Evan Eichler "
     "and Jeff Bailey</A>.");
hFreeConn(&conn);
}

void doMiddle()
/* Generate body of HTML. */
{
char *group = cgiString("g");
char *item = cgiOptionalString("i");
char title[256];

hSetDb("hg3");
hgSetDb("hg3");
seqName = cgiString("c");
winStart = cgiInt("l");
winEnd = cgiInt("r");

if (sameWord(group, "getDna"))
    {
    doGetDna();
    }
else if (sameWord(group, "hgMrna") || sameWord(group, "hgMrna2"))
    {
    doHgRna(item, FALSE);
    }
else if (sameWord(group, "hgEst"))
    {
    doHgRna(item, TRUE);
    }
else if (sameWord(group, "hgContig"))
    {
    doHgContig(item);
    }
else if (sameWord(group, "hgCover"))
    {
    doHgCover(item);
    }
else if (sameWord(group, "hgClone"))
    {
    doHgClone(item);
    }
else if (sameWord(group, "hgGold"))
    {
    doHgGold(item);
    }
else if (sameWord(group, "hgGap"))
    {
    doHgGap(item);
    }
else if (sameWord(group, "hgTet"))
    {
    doHgTet(item);
    }
else if (sameWord(group, "hgRepeat"))
    {
    doHgRepeat(item);
    }
else if (sameWord(group, "hgIsochore"))
    {
    doHgIsochore(item);
    }
else if (sameWord(group, "hgSimpleRepeat"))
    {
    doSimpleRepeat(item);
    }
else if (sameWord(group, "hgCpgIsland"))
    {
    doCpgIsland(item);
    }
else if (sameWord(group, "genieKnown"))
    {
    doKnownGene(item);
    }
else if (sameWord(group, "genieAlt"))
    {
    doGeniePred(item);
    }
else if (sameWord(group, "hgEnsGene"))
    {
    doEnsPred(item);
    }
else if (sameWord(group, "hgGenomicDups"))
    {
    doGenomicDups(item);
    }
else if (sameWord(group, "htcCloneSeq"))
    {
    htcCloneSeq(item);
    }
else if (sameWord(group, "htcCdnaAli"))
   {
   htcCdnaAli(item);
   }
else
   {
   htmlStart(group);
   printf("Sorry, clicking there doesn't do anything yet (%s).", group);
   }
htmlEnd();
}

int main(int argc, char *argv[])
{
dnaUtilOpen();
htmEmptyShell(doMiddle, NULL);
}
